# Wraps codes for PhenomHM originally from
# arXiv:1708.00404
# arXiv:1508.07250
# arXiv:1508.07253
# but now adjusted for GPU usage.

# Copyright (C) 2021 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from scipy.interpolate import CubicSpline

# import GPU stuff
try:
    from bbhx.cutils.pyPhenomHM import (
        waveform_amp_phase_wrap as waveform_amp_phase_wrap_gpu,
    )
    from bbhx.cutils.pyPhenomHM import (
        get_phenomhm_ringdown_frequencies as get_phenomhm_ringdown_frequencies_gpu,
    )
    from bbhx.cutils.pyPhenomHM import (
        get_phenomd_ringdown_frequencies as get_phenomd_ringdown_frequencies_gpu,
    )
    import cupy as cp

except (ImportError, ModuleNotFoundError) as e:
    print("No CuPy or GPU PhenomHM module.")

from bbhx.cutils.pyPhenomHM_cpu import (
    waveform_amp_phase_wrap as waveform_amp_phase_wrap_cpu,
)
from bbhx.cutils.pyPhenomHM_cpu import (
    get_phenomhm_ringdown_frequencies as get_phenomhm_ringdown_frequencies_cpu,
)
from bbhx.cutils.pyPhenomHM_cpu import (
    get_phenomd_ringdown_frequencies as get_phenomd_ringdown_frequencies_cpu,
)

from ..utils.interpolate import CubicSplineInterpolant, searchsorted2d_vec
from ..utils.constants import *
from ..waveforms.ringdownphenomd import *


class PhenomHMAmpPhase:
    """Produce PhenomHM in the amplitude and phase representation

    This class implements PhenomD and PhenomHM in a GPU-accelerated form.
    If you use this class, please cite
    `arXiv:2005.01827 <https://arxiv.org/abs/2005.01827>`_,
    `arXiv:2111.01064 <https://arxiv.org/abs/2111.01064>`_
    `arXiv:1708.00404 <https://arxiv.org/abs/1708.00404>`_,
    `arXiv:1508.07250 <https://arxiv.org/abs/1508.07250>`_, and
    `arXiv:1508.07253 <https://arxiv.org/abs/1508.07253>_.

    Args:
        use_gpu (bool, optional): If ``True``, run on the GPU.
        run_phenomd (bool, optional): If ``True``, run the PhenomD
            waveform rather than PhenomHM. Really this is the same
            as choosing ``modes=[(2,2)]`` in the PhenomHM waveform.
        mf_min (double, optional): Dimensionless minimum frequency to use when performing
            interpolation. (Default: ``1e-4``)
        mf_max (double, optional): Dimensionless maximum frequency to use when performing
            interpolation. (Default: ``6e-1``)
        initial_t_val (double, optional): Time at the start of the
            time window. This shifts the phase accordingly but does
            not shift the tf correspondence so that the response
            is still accurately reflected. (Default: ``0.0``)

    Attributes:
        allowable_modes (list): Allowed list of mode tuple pairs ``(l,m)`` for
            the chosen waveform model.
        ells_default (np.ndarray): Default values for the ``l`` index of the harmonic.
        mms_default (np.ndarray): Default values for the ``m`` index of the harmonic.
        mf_max (double): Dimensionless maximum frequency to use when performing
            interpolation.
        mf_min (double): Dimensionless minimum frequency to use when performing
            interpolation.
        phenomhm_ringdown_freqs (obj): Ringdown frequency determination in PhenomHM.
        phenomd_ringdown_freqs (obj): Ringdown frequency determination in PhenomD.
        run_phenomd (bool): If ``True``, run the PhenomD
            waveform rather than PhenomHM. Really this is the same
            as choosing ``modes=[(2,2)]`` in the PhenomHM waveform.
        y_rd (xp.ndarray): Y-values for PhenomD ringdown frequncy for Cubic Spline.
        c1_rd (xp.ndarray): Cubic Spline c1 values for PhenomD ringdown frequency.
        c2_rd (xp.ndarray): Cubic Spline c2 values for PhenomD ringdown frequency.
        c3_rd (xp.ndarray): Cubic Spline c3 values for PhenomD ringdown frequency.
        y_dm (xp.ndarray): Y-values for PhenomD damping frequncy for Cubic Spline.
        c1_dm (xp.ndarray): Cubic Spline c1 values for PhenomD damping frequency.
        c2_dm (xp.ndarray): Cubic Spline c2 values for PhenomD damping frequency.
        c3_dm (xp.ndarray): Cubic Spline c3 values for PhenomD damping frequency.
        waveform_carrier (xp.ndarray): Carrier for amplitude, phase, and tf information.

    """

    def __init__(
        self,
        use_gpu=False,
        run_phenomd=False,
        mf_min=1e-4,
        mf_max=0.6,
        initial_t_val=0.0,
    ):
        self.use_gpu = use_gpu

        self.run_phenomd = run_phenomd

        self.allowable_modes = [(2, 2), (3, 3), (4, 4), (2, 1), (3, 2), (4, 3)]

        self.ells_default = self.xp.array([2, 3, 4, 2, 3, 4], dtype=self.xp.int32)

        self.mms_default = self.xp.array([2, 3, 4, 1, 2, 3], dtype=self.xp.int32)

        # fix allowable quantities if you are running phenomd
        if self.run_phenomd:
            self.allowable_modes = [(2, 2)]

            self.ells_default = self.xp.array([2], dtype=self.xp.int32)

            self.mms_default = self.xp.array([2], dtype=self.xp.int32)

        # max/min dimensionless frequencies evaluated when
        # freqs are not given, but a length is given
        self.mf_min = mf_min
        self.mf_max = mf_max

        self.initial_t_val = initial_t_val

        # prepare the PhenomD spline info for fRD and fDM
        self._init_phenomd_fring_spline()

    @property
    def use_gpu(self) -> bool:
        """Whether to use a GPU."""
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu: bool) -> None:
        """Set ``use_gpu``."""
        assert isinstance(use_gpu, bool)
        self._use_gpu = use_gpu

    @property
    def xp(self) -> object:
        """Numpy or Cupy"""
        return cp if self.use_gpu else np

    @property
    def waveform_gen(self) -> callable:
        """C/CUDA wrapped function for computing interpolation."""
        return (
            waveform_amp_phase_wrap_gpu if self.use_gpu else waveform_amp_phase_wrap_cpu
        )

    @property
    def phenomhm_ringdown_freqs(self) -> callable:
        """C/CUDA wrapped function for computing PhenomHM Ringdown frequencies."""
        return (
            get_phenomhm_ringdown_frequencies_gpu
            if self.use_gpu
            else get_phenomhm_ringdown_frequencies_cpu
        )

    @property
    def phenomd_ringdown_freqs(self) -> callable:
        """C/CUDA wrapped function for computing PhenomD Ringdown frequencies."""
        return (
            get_phenomd_ringdown_frequencies_gpu
            if self.use_gpu
            else get_phenomd_ringdown_frequencies_cpu
        )

    @property
    def citation(self):
        """Return citations for this class"""
        return katz_citations + phenomhm_citation + phenomd_citations

    def _init_phenomd_fring_spline(self):
        """Prepare PhenomD fring and fdamp splines"""
        # setup splines
        spl_ring = CubicSpline(QNMData_a, QNMData_fring)
        spl_damp = CubicSpline(QNMData_a, QNMData_fdamp)

        # store the coefficients
        self.y_rd = self.xp.asarray(QNMData_fring).copy()
        self.c1_rd = self.xp.asarray(spl_ring.c[-2]).copy()
        self.c2_rd = self.xp.asarray(spl_ring.c[-3]).copy()
        self.c3_rd = self.xp.asarray(spl_ring.c[-4]).copy()

        self.y_dm = self.xp.asarray(QNMData_fdamp).copy()
        self.c1_dm = self.xp.asarray(spl_damp.c[-2]).copy()
        self.c2_dm = self.xp.asarray(spl_damp.c[-3]).copy()
        self.c3_dm = self.xp.asarray(spl_damp.c[-4]).copy()

    def _sanity_check_modes(self, ells, mms):
        """Make sure ell and mm combinations are available"""
        for ell, mm in zip(ells, mms):
            if (ell, mm) not in self.allowable_modes:
                raise ValueError(
                    "Requested mode [(l,m) = ({},{})] is not available. Allowable modes include {}".format(
                        ell, mm, self.allowable_modes
                    )
                )

    def _sanity_check_params(self, m1, m2, chi1z, chi2z):
        """Make sure input parameters are okay"""
        if np.any(np.abs(chi1z) > 1.0):
            raise ValueError(
                "chi1z array contains values with abs(chi1z) > 1.0 which is not allowed."
            )

        if np.any(np.abs(chi2z) > 1.0):
            raise ValueError(
                "chi2z array contains values with abs(chi2z) > 1.0 which is not allowed."
            )

        # ensure m1 > m2
        switch = m1 < m2

        m1_temp = m2[switch]
        m2[switch] = m1[switch]
        m1[switch] = m1_temp

        # adjust chi values if masses are switched
        chi1z_temp = chi2z[switch]
        chi2z[switch] = chi1z[switch]
        chi1z[switch] = chi1z_temp

        return (m1, m2, chi1z, chi2z)

    def _initialize_waveform_container(self):
        """Initialize the waveform container based on input dimensions"""
        self.waveform_carrier = self.xp.zeros(
            (self.nparams * self.length * self.num_modes * self.num_bin_all),
            dtype=self.xp.float64,
        )

    def _initialize_freqs(self, m1, m2, mms):
        """Setup frequencies when not given by user"""
        M_tot_sec = (m1 + m2) * MTSUN_SI

        # dimensionless freqs
        base_freqs = self.xp.logspace(
            self.xp.log10(self.mf_min), self.xp.log10(self.mf_max), self.length
        )

        # adjust them for each binary Mass
        # flatten to prepare for C computations
        self.freqs = (
            base_freqs[self.xp.newaxis, self.xp.newaxis, :]
            / M_tot_sec[:, self.xp.newaxis, self.xp.newaxis]
            * (mms[self.xp.newaxis, :, self.xp.newaxis])
            / 2
        ).flatten()
        # mms / 2 sets ratio versus 22 mode

    @property
    def amp(self):
        """Get the amplitude array with shape ``(num_bin_all, num_modes, length)``"""
        amps = self.waveform_carrier[
            0 * self.num_per_param : 1 * self.num_per_param
        ].reshape(self.num_bin_all, self.num_modes, self.length)
        return amps

    @property
    def phase(self):
        """Get the phase array with shape ``(num_bin_all, num_modes, length)``"""
        phase = self.waveform_carrier[
            1 * self.num_per_param : 2 * self.num_per_param
        ].reshape(self.num_bin_all, self.num_modes, self.length)
        return phase

    @property
    def tf(self):
        """Get the tf array with shape ``(num_bin_all, num_modes, length)``"""
        tf = self.waveform_carrier[
            2 * self.num_per_param : 3 * self.num_per_param
        ].reshape(self.num_bin_all, self.num_modes, self.length)
        return tf

    @property
    def freqs_shaped(self):
        """Get the freqs array with shape ``(num_bin_all, length)``"""
        return self._freqs.reshape(self.num_bin_all, self.num_modes, self.length)

    @property
    def freqs(self):
        """Get the flat freqs array"""
        return self._freqs

    @freqs.setter
    def freqs(self, f):
        """Prepare frequencies properly for C"""
        if f.ndim > 1:
            self._freqs = f.flatten()

        else:
            self._freqs = f

    def run_wave(
        self,
        m1,
        m2,
        chi1z,
        chi2z,
        distance,
        phi_ref,
        f_ref,
        t_ref,
        length,
        freqs=None,
        out_buffer=None,
        modes=None,
    ):
        """Generate PhenomHM/D waveforms

        Generate PhenomHM/PhenomD waveforms based on user given quantitites
        in the Amplitude-Phase representation.

        Args:
            m1 (double scalar or np.ndarray): Mass 1 in Solar Masses :math:`(m1 > m2)`.
            m2 (double or np.ndarray): Mass 2 in Solar Masses :math:`(m1 > m2)`.
            chi1z (double or np.ndarray): Dimensionless spin 1 (for Mass 1) in Solar Masses.
            chi2z (double or np.ndarray): Dimensionless spin 2 (for Mass 1) in Solar Masses.
            distance (double or np.ndarray): Luminosity distance in m.
            phi_ref (double or np.ndarray): Phase at ``f_ref``.
            f_ref (double or np.ndarray): Reference frequency at which ``phi_ref`` and ``t_ref`` are set.
                If ``f_ref == 0``, it will be set internally by the PhenomHM code
                to :math:`f_\\text{max} = \\text{max}(f^2A_{22}(f))`.
            t_ref (double or np.ndarray): Reference time in seconds. It is set at ``f_ref``.
            length (int): Length of the frequency array over which the waveform is created.
            freqs (1D or 2D xp.ndarray, optional): If ``None``, the class will generate the
                frequency array over which the waveform is evaluated. If 1D xp.ndarray,
                this array will be copied for all binaries evaluated. If 2D,
                it must have shape ``(num_bin_all, length)``. (Default: ``None``)
            out_buffer (xp.ndarray, optional): If ``None``, a buffer array will be created.
                If provided, it should be flattened from shape
                ``(nparams, length, num_modes, num_bin_all)``. ``nparams`` can
                be 3 if just evaluating PhenomHM/D. If using the same buffer for
                the response it must be 9. (Default: ``None``)
            modes (list, optional): Harmonic modes to use. If not given, they will
                default to those available in the waveform model. For PhenomHM:
                [(2,2), (3,3), (4,4), (2,1), (3,2), (4,3)]. For PhenomD: [(2,2)].
                (Default: ``None``)

        """

        # cast to 1D arrays if scalars
        m1 = np.atleast_1d(m1)
        m2 = np.atleast_1d(m2)
        chi1z = np.atleast_1d(chi1z)
        chi2z = np.atleast_1d(chi2z)
        distance = np.atleast_1d(distance)
        phi_ref = np.atleast_1d(phi_ref)
        f_ref = np.atleast_1d(f_ref)
        t_ref = np.atleast_1d(t_ref)

        # make sure parameters are okay and ordered so m1 > m2
        m1, m2, chi1z, chi2z = self._sanity_check_params(m1, m2, chi1z, chi2z)

        # set modes
        if modes is not None:
            ells = self.xp.asarray([ell for ell, mm in modes], dtype=self.xp.int32)
            mms = self.xp.asarray([mm for ell, mm in modes], dtype=self.xp.int32)
            self.modes = modes
            self._sanity_check_modes(ells, mms)

        else:
            self.modes = self.allowable_modes
            ells = self.ells_default
            mms = self.mms_default

        # adjust for phenomD
        if self.run_phenomd:
            ells = self.xp.asarray([2], dtype=self.xp.int32)
            mms = self.xp.asarray([2], dtype=self.xp.int32)
            self.modes = self.allowable_modes

        num_modes = len(ells)
        num_bin_all = len(m1)

        # store necessary parameters
        # here we evaluate 3 parameters: amp, phase, tf
        self.length = length
        self.num_modes = num_modes
        self.num_bin_all = num_bin_all
        self.nparams = 3
        self.num_per_param = length * num_modes * num_bin_all
        self.num_per_bin = length * num_modes

        # cast to GPU if needed
        m1 = self.xp.asarray(m1).copy()
        m2 = self.xp.asarray(m2).copy()
        chi1z = self.xp.asarray(chi1z).copy()
        chi2z = self.xp.asarray(chi2z).copy()
        distance = self.xp.asarray(distance).copy()
        phi_ref = self.xp.asarray(phi_ref).copy()
        f_ref = self.xp.asarray(f_ref).copy()

        # setup out_buffer if not given
        if out_buffer is None:
            self._initialize_waveform_container()

        else:
            # TODO: add sanity checks for buffer size like response
            self.waveform_carrier = out_buffer

        # initialize frequencies if not given
        if freqs is None:
            self._initialize_freqs(m1, m2, mms)

        elif freqs.ndim == 1:
            self.freqs = self.xp.tile(
                freqs, (self.num_bin_all, self.num_modes, 1)
            ).flatten()
        elif freqs.ndim == 3:
            self.freqs = freqs.flatten().copy()
        else:
            raise ValueError(
                "If providing freqs, must be 1D array (length,) or 3D array (length, num_modes, length)."
            )

        # convert to SI units for the mass
        m1_SI = m1 * MSUN_SI
        m2_SI = m2 * MSUN_SI

        # prepare for phenomD fring and fdamp
        self.fringdown = self.xp.zeros(self.num_modes * self.num_bin_all)
        self.fdamp = self.xp.zeros(self.num_modes * self.num_bin_all)

        # get phenomD freq info
        self.phenomd_ringdown_freqs(
            self.fringdown,
            self.fdamp,
            m1,
            m2,
            chi1z,
            chi2z,
            self.num_bin_all,
            self.y_rd,
            self.c1_rd,
            self.c2_rd,
            self.c3_rd,
            self.y_dm,
            self.c1_dm,
            self.c2_dm,
            self.c3_dm,
            dspin,
        )

        if not self.run_phenomd:
            # move phenomD results to the last entry in the array after
            # phenomhm frequencies
            append_phenomd_frd = self.fringdown[: self.num_bin_all].copy()
            append_phenomd_fdm = self.fdamp[: self.num_bin_all].copy()

            # get phenomhm frequencies
            self.phenomhm_ringdown_freqs(
                self.fringdown,
                self.fdamp,
                m1,
                m2,
                chi1z,
                chi2z,
                ells,
                mms,
                self.num_modes,
                self.num_bin_all,
            )

            # this adds the phenomD frequencies to keep everything consistent
            self.fringdown = (
                self.xp.concatenate(
                    [
                        self.fringdown.reshape(-1, num_modes),
                        self.xp.array([append_phenomd_frd]).T,
                    ],
                    axis=1,
                )
                .flatten()
                .copy()
            )
            self.fdamp = (
                self.xp.concatenate(
                    [
                        self.fdamp.reshape(-1, num_modes),
                        self.xp.array([append_phenomd_fdm]).T,
                    ],
                    axis=1,
                )
                .flatten()
                .copy()
            )

        # inside this code, t_ref is zero and phi_ref is zero
        self.waveform_gen(
            self.waveform_carrier,
            ells,
            mms,
            self.freqs,
            m1_SI,
            m2_SI,
            chi1z,
            chi2z,
            distance,
            f_ref,
            num_modes,
            length,
            num_bin_all,
            self.fringdown,
            self.fdamp,
            self.run_phenomd,
        )

        # adjust phases based on shift from t_ref
        # do this inplace
        temp = (
            self.freqs.reshape(self.num_bin_all, self.num_modes, -1)
            * self.xp.asarray(
                t_ref[:, self.xp.newaxis, self.xp.newaxis] - self.initial_t_val
            )
            * 2
            * np.pi
        )

        # phases = self.waveform_carrier[1]  (waveform carrier is flat)
        self.waveform_carrier[
            1 * self.num_per_param : 2 * self.num_per_param
        ] += temp.flatten()

        # adjust t-f for shift of t_ref
        # t_ref array = self.waveform_carrier[2] (waveform carrier is flat)
        self.waveform_carrier[
            2 * self.num_per_param : 3 * self.num_per_param
        ] += self.xp.tile(
            self.xp.asarray(t_ref)[:, None, None], (1, self.num_modes, self.length)
        ).flatten()

    def __call__(self, *args, Tobs=None, direct=False, **kwargs):

        self.run_wave(*args, **kwargs)

        # interpolating. Need to adjust frequency bounds
        if not direct:
            # TODO: must check if this is accurate enough!!!!!!!

            # PURPOSE: NEED TO REMOVE FREQUENCIES WHERE ORBIT INFO DOES NOT EXIST
            # get start of available information

            fix = self.tf.min(axis=-1) < 0.0
            f_min = self.freqs_shaped.min(axis=-1)

            if self.xp.any(fix):
                windows_of_interest = self.xp.argwhere(
                    (0.0 < self.tf[fix, 1:]) & (0.0 > self.tf[fix, :-1])
                )

                if self.xp.unique(windows_of_interest, axis=0).shape[0] != fix.sum():
                    raise ValueError(
                        "The search for a window at x_min did not run correctly. If t_ref is close to zero, the time-frequency correspondence may be crossing zero twice."
                    )

                # linear interp for minimum frequency
                segment_inds = windows_of_interest[:, 1]

                y1 = self.xp.take_along_axis(
                    self.freqs_shaped[fix], segment_inds[:, None], axis=-1
                )[:, 0]
                y2 = self.xp.take_along_axis(
                    self.freqs_shaped[fix], segment_inds[:, None] + 1, axis=-1
                )[:, 0]
                x1 = self.xp.take_along_axis(
                    self.tf[fix], segment_inds[:, None], axis=-1
                )[:, 0]
                x2 = self.xp.take_along_axis(
                    self.tf[fix], segment_inds[:, None] + 1, axis=-1
                )[:, 0]

                m = (y2 - y1) / (x2 - x1)
                b = y1

                f_min_new = m * (0.0 - x1) + b
                f_min[fix] = f_min_new

            if Tobs is None:
                # will default to tf bound
                Tobs = self.xp.full((self.num_bin_all, self.num_modes), 1e300)
            elif isinstance(Tobs, float):
                Tobs = self.xp.full((self.num_bin_all, self.num_modes), Tobs)
            elif isinstance(Tobs, np.ndarray) and Tobs.ndim == 1:
                assert Tobs.shape[0] == self.num_bin_all
                Tobs = self.xp.repeat(
                    self.xp.asarray(Tobs)[:, None], self.num_modes, axis=-1
                )
            elif isinstance(Tobs, np.ndarray) and Tobs.ndim == 2:
                Tobs = self.xp.asarray(Tobs)
                assert Tobs.shape == (self.num_bin_all, self.num_modes)
            else:
                raise ValueError("Tobs entered incorrectly.")

            fix = self.tf.max(axis=-1) > Tobs
            f_max = self.freqs_shaped.max(axis=-1)

            if self.xp.any(fix):
                windows_of_interest = self.xp.argwhere(
                    (Tobs[fix][:, None] < self.tf[fix, 1:])
                    & (Tobs[fix][:, None] > self.tf[fix, :-1])
                )

                if self.xp.unique(windows_of_interest, axis=0).shape[0] != fix.sum():
                    raise ValueError(
                        "The search for a window at x_max did not run correctly. If t_ref is close to the observation time, the time-frequency correspondence may be crossing zero twice."
                    )

                # linear interp for minimum frequency
                segment_inds = windows_of_interest[:, 1]

                y1 = self.xp.take_along_axis(
                    self.freqs_shaped[fix], segment_inds[:, None], axis=-1
                )[:, 0]
                y2 = self.xp.take_along_axis(
                    self.freqs_shaped[fix], segment_inds[:, None] + 1, axis=-1
                )[:, 0]
                x1 = self.xp.take_along_axis(
                    self.tf[fix], segment_inds[:, None], axis=-1
                )[:, 0]
                x2 = self.xp.take_along_axis(
                    self.tf[fix], segment_inds[:, None] + 1, axis=-1
                )[:, 0]

                m = (y2 - y1) / (x2 - x1)
                b = y1

                f_max_new = m * (Tobs[fix] - x1) + b

                f_max[fix] = f_max_new

            new_freqs = self.xp.logspace(
                self.xp.log10(f_min), self.xp.log10(f_max), self.length, axis=-1
            )

            tmp_kwargs = kwargs.copy()
            tmp_kwargs["freqs"] = new_freqs
            self.run_wave(*args, **tmp_kwargs)



import shutil


fps_cu_to_cpp = [
    "PhenomHM",
    "Response",
    "Interpolate",
    "WaveformBuild",
    "Likelihood",
]
fps_pyx = ["phenomhm", "response", "interpolate", "waveformbuild", "likelihood"]

for fp in fps_cu_to_cpp:
    shutil.copy("bbhx/cutils/src/" + fp + ".cu", "bbhx/cutils/src/" + fp + ".cpp")

for fp in fps_pyx:
    shutil.copy("bbhx/cutils/src/" + fp + ".pyx", "bbhx/cutils/src/" + fp + "_cpu.pyx")


# setup version file
with open("README.md", "r") as fh:
    lines = fh.readlines()

for line in lines:
    if line.startswith("Current Version"):
        version_string = line.split("Current Version: ")[1].split("\n")[0]

with open("bbhx/_version.py", "w") as f:
    f.write("__version__ = '{}'".format(version_string))


initial_text_for_constants_file = """
# Collection of citations for modules in bbhx package

# Copyright (C) 2020 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""

initial_text_for_constants_file2 = """

Constants
*******************************

The module :code:`bbhx.utils.constants` houses the constants used throughout the package.

Constants list:

"""

fp_out_name = "bbhx/utils/constants.py"
fp_out_name2 = "docs/source/user/constants.rst"
fp_in_name = "bbhx/cutils/include/constants.h"

# develop few.utils.constants.py
with open(fp_out_name, "w") as fp_out:
    with open(fp_out_name2, "w") as fp_out2:
        fp_out.write(initial_text_for_constants_file)
        fp_out2.write(initial_text_for_constants_file2)
        with open(fp_in_name, "r") as fp_in:
            lines = fp_in.readlines()
            for line in lines:
                if len(line.split()) == 3:
                    if line.split()[0] == "#define":
                        try:
                            _ = float(line.split()[2])
                            string_out = (
                                line.split()[1] + " = " + line.split()[2] + "\n"
                            )
                            fp_out.write(string_out)
                            fp_out2.write(f"\t* {string_out}")

                        except ValueError as e:
                            continue

# for fp in fps_cu_to_cpp:
#     os.remove("src/" + fp + ".cpp")

# for fp in fps_pyx:
#     os.remove("src/" + fp + "_cpu.pyx")



# Collection of citations for modules in bbhx package

# Copyright (C) 2020 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
:code:`bbhx.utils.citations`:

This module is used to collect citations for all modules in the package. This
module is then imported to add citations to module classes using their :code:`citation`
attribute.
"""

katz_citations = """
@article{Katz:2020hku,
    author = "Katz, Michael L. and Marsat, Sylvain and Chua, Alvin J. K. and Babak, Stanislav and Larson, Shane L.",
    title = "{GPU-accelerated massive black hole binary parameter estimation with LISA}",
    eprint = "2005.01827",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.102.023033",
    journal = "Phys. Rev. D",
    volume = "102",
    number = "2",
    pages = "023033",
    year = "2020"
}

@article{Katz:2021uax,
    author = "Katz, Michael L.",
    title = "{A fully-automated end-to-end pipeline for massive black hole binary signal extraction from LISA data}",
    eprint = "2111.01064",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    month = "11",
    year = "2021"
}

@article{michael katz_2021,
    title={mikekatz04/BBHx: First official public release},
    DOI={10.5281/zenodo.5730688},
    abstractNote={No description provided.},
    publisher={Zenodo},
    author={Michael Katz},
    year={2021},
    month={Nov}
}
"""

marsat_1 = """
@article{Marsat:2018oam,
    author = "Marsat, Sylvain and Baker, John G.",
    title = "{Fourier-domain modulations and delays of gravitational-wave signals}",
    eprint = "1806.10734",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    month = "6",
    year = "2018"
}
"""

marsat_2 = """
@article{Marsat:2020rtl,
    author = "Marsat, Sylvain and Baker, John G. and Dal Canton, Tito",
    title = "{Exploring the Bayesian parameter estimation of binary black holes with LISA}",
    eprint = "2003.00357",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.103.083011",
    journal = "Phys. Rev. D",
    volume = "103",
    number = "8",
    pages = "083011",
    year = "2021"
}
"""

phenomhm_citation = """
@article{London2018,
    archivePrefix = {arXiv},
    arxivId = {gr-qc/1708.00404},
    author = {London, Lionel and Khan, Sebastian and Fauchon-Jones, Edward and Garc\'ia, Cecilio and Hannam, Mark and Husa, Sascha and Jim{\'{e}}nez-Forteza, Xisco and Kalaghatgi, Chinmay and Ohme, Frank and Pannarale, Francesco},
    doi = {10.1103/PhysRevLett.120.161102},
    eprint = {1708.00404},
    journal = {\prl},
    keywords = {General Relativity and Quantum Cosmology},
    month = {apr},
    number = {16},
    pages = {161102},
    primaryClass = {gr-qc},
    title = {{First Higher-Multipole Model of Gravitational Waves from Spinning and Coalescing Black-Hole Binaries}},
    volume = {120},
    year = {2018}
}
"""

phenomd_citations = """
@article{Husa2016,
    archivePrefix = {arXiv},
    arxivId = {gr-qc/1508.07250},
    author = {Husa, S and Khan, S and Hannam, M and P{\"{u}}rrer, M and Ohme, F and Forteza, X.$\sim$J. and Boh{\'{e}}, A},
    doi = {10.1103/PhysRevD.93.044006},
    eprint = {1508.07250},
    journal = {\prd},
    month = {feb},
    number = {4},
    pages = {44006},
    primaryClass = {gr-qc},
    title = {{Frequency-domain gravitational waves from nonprecessing black-hole binaries. I. New numerical waveforms and anatomy of the signal}},
    volume = {93},
    year = {2016}
}


@article{Khan2016,
    archivePrefix = {arXiv},
    arxivId = {gr-qc/1508.07253},
    author = {Khan, S and Husa, S and Hannam, M and Ohme, F and P{\"{u}}rrer, M and Forteza, X.$\sim$J. and Boh{\'{e}}, A},
    doi = {10.1103/PhysRevD.93.044007},
    eprint = {1508.07253},
    journal = {\prd},
    month = {feb},
    number = {4},
    pages = {44007},
    primaryClass = {gr-qc},
    title = {{Frequency-domain gravitational waves from nonprecessing black-hole binaries. II. A phenomenological model for the advanced detector era}},
    volume = {93},
    year = {2016}
}
"""

Cornish_Heterodyning = """
@article{Cornish:2021lje,
    author = "Cornish, Neil J.",
    title = "{Heterodyned likelihood for rapid gravitational wave parameter inference}",
    eprint = "2109.02728",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.104.104054",
    journal = "Phys. Rev. D",
    volume = "104",
    number = "10",
    pages = "104054",
    year = "2021"
}
"""

Rel_Bin_citation = """
@article{Zackay:2018qdy,
    author = "Zackay, Barak and Dai, Liang and Venumadhav, Tejaswi",
    title = "{Relative Binning and Fast Likelihood Evaluation for Gravitational Wave Parameter Estimation}",
    eprint = "1806.08792",
    archivePrefix = "arXiv",
    primaryClass = "astro-ph.IM",
    month = "6",
    year = "2018"
}
"""



# Fast Likelihood functions

# Copyright (C) 2021 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

try:
    import cupy as cp
    from bbhx.cutils.pyLikelihood import hdyn_wrap as hdyn_wrap_gpu
    from bbhx.cutils.pyLikelihood import direct_like_wrap as direct_like_wrap_gpu
    from bbhx.cutils.pyLikelihood import prep_hdyn as prep_hdyn_gpu

except (ImportError, ModuleNotFoundError) as e:
    print("No CuPy")
    import numpy as cp

from bbhx.cutils.pyLikelihood_cpu import prep_hdyn as prep_hdyn_cpu
from bbhx.cutils.pyLikelihood_cpu import hdyn_wrap as hdyn_wrap_cpu
from bbhx.cutils.pyLikelihood_cpu import direct_like_wrap as direct_like_wrap_cpu

from bbhx.utils.constants import *

from lisatools.sensitivity import SensitivityMatrix, AET1SensitivityMatrix


class Likelihood:
    """Fast Base Likelihood Class for MBHBs

    This class computes the graitational wave Likelihood as a direct sum over
    frequecy bins. It only sums over the frequencies where the MBHB signal
    exists. Therefore, larger mass waveforms are faster because there are less
    frequencies. This class computes:
    :math:`\\mathcal{L}\\propto-1/2\\langle d-h|d-h\\rangle=-1/2\\left(\\langle d|d\\rangle \\langle h|h\\rangle - 2\\langle d|h\\rangle\\right)`.

    This class has GPU capability.

    Args:
        template_gen (obj): Waveform generation class that returns a tuple of
            (list of template arrays, start indices, lengths). See
            :class:`bbhx.waveform.BBHWaveformFD` for more information on this
            return type.
        data_freqs (double xp.ndarray): Frequencies for the data stream. ``data_freqs``
            should be a numpy (cupy) array if running on the CPU (GPU).
        data_channels (complex128 xp.ndarray): Data stream. 2D array of shape: ``(3, len(data_freqs))``.
            It is assumed there are 3 channels. ``data_channels``
            should be a numpy (cupy) array if running on the CPU (GPU).
        psd (double xp.ndarray): Power Spectral Density in the noise:math:`S_n(f)`.
            2D array of shape: ``(3, len(data_freqs))``.
            It is assumed there are 3 channels. ``psd``
            should be a numpy (cupy) array if running on the CPU (GPU).
        use_gpu (bool, optional): If ``True``, use GPU.

    Attributes:
        d_d (double): :math:`\\langle d|d\\rangle` inner product value.
        data_channels (complex128 np.ndarray): Data stream. 1D flattened array
            of shape: ``(3, len(data_freqs))``. **Note** ``data_channels`` should
            be multiplied by ``psd`` before input into this class.
        data_freqs (double np.ndarray): Frequencies for the data stream (1D).
        data_stream_length (int): Length of data.
        noise_factors (double xp.ndarray): :math:`\\sqrt{\\frac{\\Delta f}{S_n(f)}}`.
            1D flattened array of shape: ``(3, len(data_freqs))``.
        psd (double xp.ndarray): Power Spectral Density in the noise:math:`S_n(f)`.
            1D flattened array of shape: ``(3, len(data_freqs))``.
        template_gen (obj): Waveform generation class that returns a tuple of
            (list of template arrays, start indices, lengths). See
            :class:`bbhx.waveform.BBHWaveformFD` for more information on this
            return type.
        phase_marginalize (bool): If ``True``, compute the phase-marginalized
            log-Likelihood (and snr if ``return_extracted_snr==True``).
        return_extracted_snr (bool): Return the snr in addition to the Likeilihood.


    """

    def __init__(
        self,
        template_gen,
        data_freqs,
        data_channels,
        psd,
        use_gpu=False,
    ):

        self.use_gpu = use_gpu

        # store required information
        self.data_freqs = data_freqs

        try:
            data_freqs_cpu = data_freqs.get()
        except AttributeError:
            data_freqs_cpu = data_freqs

        # will not store psd or delta_f on GPU if being used
        try:
            psd = psd.get()
        except AttributeError:
            pass

        self.psd = np.asarray(psd)
        self.delta_f = np.zeros_like(self.psd)
        self.delta_f[:, 1:] = np.diff(data_freqs_cpu)
        self.delta_f[:, 0] = self.delta_f[:, 1]

        self.noise_factors = self.xp.asarray(
            np.sqrt(1.0 / psd * self.delta_f)
        ).flatten()

        # store and adjust data_channels
        self.data_channels = self.noise_factors * self.xp.asarray(
            data_channels.flatten()
        )

        self.waveform_gen = template_gen
        self.data_stream_length = len(data_freqs)

        # assumes data_channels is already factored by psd
        self.d_d = (
            4 * self.xp.sum((self.data_channels.conj() * self.data_channels)).real
        ).item()

    @property
    def use_gpu(self) -> bool:
        """Whether to use a GPU."""
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu: bool) -> None:
        """Set ``use_gpu``."""
        assert isinstance(use_gpu, bool)
        self._use_gpu = use_gpu

    @property
    def like_gen(self):
        """Likelihood for either GPU or CPU."""
        like_gen = direct_like_wrap_gpu if self.use_gpu else direct_like_wrap_cpu
        return like_gen

    @property
    def xp(self):
        """Cupy or Numpy"""
        xp = cp if self.use_gpu else np
        return xp

    @property
    def citation(self):
        return katz_citations

    def get_ll(
        self,
        params,
        return_extracted_snr=False,
        phase_marginalize=False,
        **waveform_kwargs
    ):
        """Compute the log-Likelihood

        params (double np.ndarray): Parameters for evaluating log-Likelihood.
            ``params.shape=(num_params,)`` if 1D or
            ``params.shape=(num_params, num_bin_all)`` if 2D for more than
            one binary.
        return_extracted_snr (bool, optional): If ``True``, return
            :math:`\\langle d|h\\rangle\\ / \\sqrt{\\langle h|h\\rangle}` as a second entry
            of the return array. This produces a return array of
            ``xp.array([log likelihood, snr]).T``. If ``False``, just return
            the log-Likelihood array.
        phase_marginalize (bool, optional): If ``True``, compute the phase-marginalized
            log-Likelihood (and snr if ``return_extracted_snr==True``).
        **waveform_kwargs (dict, optional): Keyword arguments for waveform
            generator.

        Returns:
            np.ndarray: log-Likelihoods or ``np.array([log-Likelihoods, snr]).T``

        """

        # store info
        self.phase_marginalize = phase_marginalize
        self.return_extracted_snr = return_extracted_snr

        # setup kwargs properly
        waveform_kwargs["freqs"] = self.data_freqs
        waveform_kwargs["fill"] = False
        waveform_kwargs["direct"] = False

        # get information from waveform generators
        templateChannels, inds_start, ind_lengths = self.waveform_gen(
            *params, **waveform_kwargs
        )

        # get flattened templates to remove channel dimension
        templateChannels = [tc.flatten() for tc in templateChannels]

        # adjust for cupy vs numpy
        try:
            templateChannels_ptrs = np.asarray(
                [tc.data.ptr for tc in templateChannels], dtype=np.int64
            )
        except AttributeError:
            templateChannels_ptrs = np.asarray(
                [tc.__array_interface__["data"][0] for tc in templateChannels],
                dtype=np.int64,
            )

        # initialize inner product info
        self.d_h = np.zeros(self.waveform_gen.num_bin_all, dtype=self.xp.complex128)
        self.h_h = np.zeros(self.waveform_gen.num_bin_all, dtype=self.xp.complex128)

        self.like_gen(
            self.d_h,
            self.h_h,
            self.data_channels,
            self.noise_factors,
            templateChannels_ptrs,
            inds_start,
            ind_lengths,
            self.data_stream_length,
            self.waveform_gen.num_bin_all,
        )

        # phase marginalize in d_h term
        d_h_temp = self.d_h if not self.phase_marginalize else self.xp.abs(self.d_h)
        out = -1 / 2 * (self.d_d + self.h_h - 2 * d_h_temp).real
        # get out of cupy if needed
        try:
            out = out.get()

        except AttributeError:
            pass

        if self.return_extracted_snr:
            return np.array([out, d_h_temp.real / np.sqrt(self.h_h.real)]).T
        else:
            return out


class HeterodynedLikelihood:
    """Compute the Heterodyned log-Likelihood

    Heterdyning involves separating the fast and slow evolutions when comparing
    signals. This involves comparing a reference template to the data stream and determining
    various quantities at the full frequency resolution of the data stream. Then, during
    online computation, the log-Likelihood is determined by computing a new waveform
    on a sparse frequency grid and comparing it to the reference waveform on the same
    sparse grid. The practical aspect is the computation of a smaller number of frequency
    points which lowers the required  memory and increases the speed of the computation.
    More information on the general method can be found in
    `arXiv:2109.02728 <https://arxiv.org/abs/2109.02728>`_. We implement the method
    as described in `arXiv:1806.08792 <https://arxiv.org/abs/1806.08792>`_.

    This class also works with higher order harmonic modes, but it has not been tested extensivally.
    It only does a direct summation over the modes rather than heterodyning per mode. So, it is less reliable,
    but in practice it produces a solid posterior distribution.

    This class has GPU capabilities.

    Args:
        template_gen (obj): Waveform generation class that returns a tuple of
            (list of template arrays, start indices, lengths). See
            :class:`bbhx.waveform.BBHWaveformFD` for more information on this
            return type.
        data_freqs (double xp.ndarray): Frequencies for the data stream. ``data_freqs``
            should be a numpy (cupy) array if running on the CPU (GPU).
        data_channels (complex128 xp.ndarray): Data stream. 2D array of shape: ``(3, len(data_freqs))``.
            It is assumed there are 3 channels. ``data_channels``
            should be a numpy (cupy) array if running on the CPU (GPU).
        reference_template_params (np.ndarray): Parameters for the reference template for
            ``template_gen``.
        template_gen_kwargs (dict, optional): Keywords arguments for generating the
            template with ``template_gen``. It will automatically add/change ``direct``
            and ``compress`` keyword arguments to ``True``. The keyword argument
            ``squeeze`` is automatically set to ``True`` for the initial setup and then
            automatically set to ``False`` for online computations. If you so choose (not
            recomended), you can change these kwargs for online running using the
            ``**waveform_kwargs`` setup in the ``self.get_ll`` method. However,
            if you include in this adjustment ``direct``, ``compress``, or ``squeeze``
            will still automatically be overwritten.
            (Default: ``{}``)
        reference_gen_kwargs (dict, optional): Keywords arguments for generating the
            reference template with ``template_gen``. It will automatically add/change
            ``fill``, ``compress``, and ``squeeze`` keyword arguments to ``True``.
            These waveforms can be produced without interpolation with ``direct = True``.
            If ``length`` keyword argument is given and ``direct=False``, the
            waveform will be interpolated. If ``length`` and ``direct`` are not given,
            it will interpolate the signal with ``length=8096``.
            (Default: ``{}``)
        sens_mat (SensitivityMatrix, optional): :class:`SensitivityMatrix` object representing the AET channels.
            If ``None``, defaults to class:`AET1SensitivityMatrix`. (default: ``None``)
        use_gpu (bool, optional): If ``True``, use GPU.

    Attributes:
        reference_d_d (double): :math:`\\langle d|d\\rangle` inner product value.
        reference_h_h(double): :math:`\\langle h|h\\rangle` inner product value
            for the reference template.
        reference_d_h (double): :math:`\\langle d|h\\rangle` inner product value
            for the reference template.
        reference_ll (double): log-Likelihood value for the reference template.
        hdyn_d_h (complex128 xp.ndarray): Heterodyned :math:`\\langle d|h\\rangle`
            inner product values for the test templates.
        hdyn_h_h (complex128 xp.ndarray): Heterodyned :math:`\\langle d|h\\rangle`
            inner product values for the test templates.
        h0_sparse (xp.ndarray): Array with sparse waveform for reference parameters.
        h_sparse (xp.ndarray): Array with sparse waveform for test parameters.
        d (complex128 xp.ndarray): Data stream. 1D flattened array
            of shape: ``(3, len(data_freqs))``.
        data_stream_length (int): Length of data.
        data_constants (xp.ndarray): Flattened array container holding all heterodyning
            constants needed: A0, A1, B0, B1.
        f_dense (xp.ndarray): Frequencies for the data stream (1D).
        freqs (xp.ndarray): Frequencies for sparse arrays.
        f_m (xp.ndarray): Frequency of mid-point in each sparse bin.
        length_f_het (int): Length of sparse array.
        psd (double xp.ndarray): :math:`\\sqrt{\\frac{\\Delta f}{S_n(f)}}`.
            1D flattened array of shape: ``(3, len(data_freqs))``.
        return_extracted_snr (bool): Return the snr in addition to the Likeilihood.
        phase_marginalize (bool): If ``True``, compute the phase-marginalized
            log-Likelihood (and snr if ``return_extracted_snr==True``).

    """

    def __init__(
        self,
        template_gen,
        data_freqs,
        data_channels,
        reference_template_params,
        length_f_het,
        template_gen_kwargs={},
        reference_gen_kwargs={},
        sens_mat=None,
        use_gpu=False,
    ):

        # store all input information
        self.template_gen = template_gen
        self.f_dense = data_freqs
        self.d = data_channels
        self.length_f_het = length_f_het

        # direct based on GPU usage
        self.use_gpu = use_gpu

        self.sens_mat = sens_mat

        # calculate all quantites related to the reference template
        self.init_heterodyne_info(
            reference_template_params,
            template_gen_kwargs=template_gen_kwargs,
            reference_gen_kwargs=reference_gen_kwargs,
        )

    @property
    def use_gpu(self) -> bool:
        """Whether to use a GPU."""
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu: bool) -> None:
        """Set ``use_gpu``."""
        assert isinstance(use_gpu, bool)
        self._use_gpu = use_gpu

    @property
    def sens_mat(self):
        """Sensitivity Matrix"""
        return self._sens_mat

    @sens_mat.setter
    def sens_mat(self, sens_mat):
        if sens_mat is None:
            _f_not_needed = np.logspace(-5, -1, 1000)
            sens_mat = AET1SensitivityMatrix(_f_not_needed)
        assert isinstance(sens_mat, SensitivityMatrix)
        self._sens_mat = sens_mat

    @property
    def like_gen(self):
        """C function on GPU/CPU"""
        like_gen = hdyn_wrap_gpu if self.use_gpu else hdyn_wrap_cpu
        return like_gen

    @property
    def xp(self):
        """Numpy or Cupy"""
        xp = cp if self.use_gpu else np
        return xp

    @property
    def citation(self):
        """Citations for this class"""
        return katz_citations + Cornish_Heterodyning + Rel_Bin_citation

    def init_heterodyne_info(
        self,
        reference_template_params,
        template_gen_kwargs={},
        reference_gen_kwargs={},
    ):
        """Prepare all information for Heterdyning

        Args:
            reference_template_params (np.ndarray): Parameters for the reference template for
                ``template_gen``.
            template_gen_kwargs (dict, optional): Keywords arguments for generating the
                template with ``template_gen``. It will automatically add/change ``direct``
                and ``compress`` keyword arguments to ``True``. The keyword argument
                ``squeeze`` is automatically set to ``True`` for the initial setup and then
                automatically set to ``False`` for online computations. If you so choose (not
                recomended), you can change these kwargs for online running using the
                ``**waveform_kwargs`` setup in the ``self.get_ll`` method. However,
                if you include in this adjustment ``direct``, ``compress``, or ``squeeze``
                will still automatically be overwritten.
                (Default: ``{}``)
            reference_gen_kwargs (dict, optional): Keywords arguments for generating the
                reference template with ``template_gen``. It will automatically add/change
                ``fill``, ``compress``, and ``squeeze`` keyword arguments to ``True``.
                These waveforms can be produced without interpolation with ``direct = True``.
                If ``length`` keyword argument is given and ``direct=False``, the
                waveform will be interpolated. If ``length`` and ``direct`` are not given,
                it will interpolate the signal with ``length=8096``.
                (Default: ``{}``)

        """

        # add the necessary kwargs for the initial template generation process.
        template_gen_kwargs["squeeze"] = True
        template_gen_kwargs["compress"] = True
        template_gen_kwargs["direct"] = True

        # need to be just outside the values of the waveform
        minF = self.f_dense.min() * 0.999999999999
        maxF = self.f_dense.max() * 1.000000000001

        if minF == 0.0:
            minF = 1e-6

        reference_gen_kwargs["squeeze"] = True
        reference_gen_kwargs["compress"] = True
        reference_gen_kwargs["fill"] = True

        # setup for reference template generation
        if "direct" in reference_gen_kwargs and reference_gen_kwargs["direct"] is True:
            pass
        else:
            reference_gen_kwargs["direct"] = False
            if "length" not in reference_gen_kwargs:
                reference_gen_kwargs["length"] = 8096

        # Heterodyning grid
        freqs = self.xp.logspace(
            self.xp.log10(minF), self.xp.log10(maxF), self.length_f_het
        )

        # generate dense reference template
        h0 = self.template_gen(
            *reference_template_params, freqs=self.f_dense, **reference_gen_kwargs
        )[0]

        # generate sparse reference template
        h0_temp = self.template_gen(
            *reference_template_params, freqs=freqs, **template_gen_kwargs
        )[0]

        # get rid of places where freqs are zero and narrow boundaries
        freqs_keep = freqs[~(self.xp.abs(h0_temp) == 0.0)]

        freqs = self.xp.logspace(
            self.xp.log10(freqs_keep[0]),
            self.xp.log10(freqs_keep[-1]),
            self.length_f_het,
        )

        # regenerate at only non-zero values of the waveform
        self.h0_sparse = self.template_gen(
            *reference_template_params, freqs=freqs, **template_gen_kwargs
        )[self.xp.newaxis, :, :]

        # find which frequencies in the dense array in contained in the sparse array
        inds = (self.f_dense >= freqs[0]) & (self.f_dense <= freqs[-1])

        # narrow the dense arrays to these indices
        self.f_dense = self.f_dense[inds]
        self.d = self.d[:, inds]
        h0 = h0[:, inds]

        # find which sparse array bins the dense frequencies fit into
        bins = self.xp.searchsorted(freqs, self.f_dense, "right") - 1

        # get frequency at middle of the bin
        f_m = (freqs[1:] + freqs[:-1]) / 2

        df = self.f_dense[1] - self.f_dense[0]

        # should be on CPU for sensitivity computation
        try:
            f_n_host = self.f_dense.get()
        except AttributeError:
            f_n_host = self.f_dense

        self.sens_mat.update_frequency_arr(f_n_host)

        # compute sensitivity at dense frequencies
        S_n = self.xp.asarray([self.sens_mat[0], self.sens_mat[1], self.sens_mat[2]])

        # compute the individual frequency contributions to A0, A1 (see paper)
        A0_flat = 4 * (h0.conj() * self.d) / S_n * df
        A1_flat = 4 * (h0.conj() * self.d) / S_n * df * (self.f_dense - f_m[bins])

        # compute the individual frequency contributions to B0, B1 (see paper)
        B0_flat = 4 * (h0.conj() * h0) / S_n * df
        B1_flat = 4 * (h0.conj() * h0) / S_n * df * (self.f_dense - f_m[bins])

        # initialize containers for sparse sums of A0, A1, B0, B1
        A0_in = self.xp.zeros((3, self.length_f_het), dtype=np.complex128)
        A1_in = self.xp.zeros_like(A0_in)
        B0_in = self.xp.zeros_like(A0_in)
        B1_in = self.xp.zeros_like(A0_in)

        for ind in self.xp.unique(bins[:-1]):
            inds_keep = bins == ind

            # TODO: check this
            inds_keep[-1] = False

            # +1 allows for zero as the first entry (for C compatibility)
            A0_in[:, ind + 1] = self.xp.sum(A0_flat[:, inds_keep], axis=1)
            A1_in[:, ind + 1] = self.xp.sum(A1_flat[:, inds_keep], axis=1)
            B0_in[:, ind + 1] = self.xp.sum(B0_flat[:, inds_keep], axis=1)
            B1_in[:, ind + 1] = self.xp.sum(B1_flat[:, inds_keep], axis=1)

        # compute stored array of all coefficients
        self.data_constants = self.xp.concatenate(
            [A0_in.flatten(), A1_in.flatten(), B0_in.flatten(), B1_in.flatten()]
        )

        # reference quantities
        self.reference_d_d = self.xp.sum(4 * (self.d.conj() * self.d) / S_n * df).real

        self.reference_h_h = self.xp.sum(B0_flat).real

        self.reference_d_h = self.xp.sum(A0_flat).real

        self.reference_ll = (
            -1 / 2 * (self.reference_d_d + self.reference_h_h - 2 * self.reference_d_h)
        )

        # sparse frequencies
        self.freqs = freqs

        # middle bin frequencies
        self.f_m = f_m

        # prepare kwargs for online evaluation
        template_gen_kwargs["squeeze"] = False
        self.template_gen_kwargs = template_gen_kwargs

    def get_ll(
        self,
        params,
        return_extracted_snr=False,
        phase_marginalize=False,
        **waveform_kwargs
    ):
        """Compute the log-Likelihood

        params (double np.ndarray): Parameters for evaluating log-Likelihood.
            ``params.shape=(num_params,)`` if 1D or
            ``params.shape=(num_params, num_bin_all)`` if 2D for more than
            one binary.
        return_extracted_snr (bool, optional): If ``True``, return
            :math:`\\langle d|h\\rangle\\ / \\sqrt{\\langle h|h\\rangle}` as a second entry
            of the return array. This produces a return array of
            ``xp.array([log likelihood, snr]).T``. If ``False``, just return
            the log-Likelihood array.
        phase_marginalize (bool, optional): If ``True``, compute the phase-marginalized
            log-Likelihood (and snr if ``return_extracted_snr==True``).
        **waveform_kwargs (dict, optional): Keyword arguments for waveform
            generator. Some may be overwritten. See the main class docstring.

        Returns:
            np.ndarray: log-Likelihoods or ``np.array([log-Likelihoods, snr]).T``

        """

        # store info
        self.phase_marginalize = phase_marginalize
        self.return_extracted_snr = return_extracted_snr

        # setup kwargs
        all_kwargs_keys = list(
            set(list(waveform_kwargs.keys()) + list(self.template_gen_kwargs.keys()))
        )

        for key in all_kwargs_keys:
            if key in ["direct", "compress", "squeeze"]:
                waveform_kwargs[key] = self.template_gen_kwargs[key]
            else:
                if key in self.template_gen_kwargs:
                    waveform_kwargs[key] = waveform_kwargs.get(
                        key, self.template_gen_kwargs[key]
                    )

        # set the frequencies at which the waveform is evaluated
        waveform_kwargs["freqs"] = self.freqs

        # compute the new sparse template
        self.h_sparse = self.template_gen(*params, **waveform_kwargs)

        # compute complex residual
        r = self.h_sparse / self.h0_sparse

        # initialize container for inner products term
        self.hdyn_d_h = self.xp.zeros(
            self.template_gen.num_bin_all, dtype=self.xp.complex128
        )
        self.hdyn_h_h = self.xp.zeros(
            self.template_gen.num_bin_all, dtype=self.xp.complex128
        )

        # adjust the residuals for entry into C
        residuals_in = r.transpose((2, 1, 0)).flatten()

        self.like_gen(
            self.hdyn_d_h,
            self.hdyn_h_h,
            residuals_in,
            self.data_constants,
            self.freqs,
            self.template_gen.num_bin_all,
            len(self.freqs),
            3,
        )

        # if phase marginalize
        d_h_temp = (
            self.hdyn_d_h if not self.phase_marginalize else self.xp.abs(self.hdyn_d_h)
        )

        # log-Likelihood
        out = -1 / 2.0 * (self.reference_d_d + self.hdyn_h_h - 2 * d_h_temp).real

        # move to CPU if needed
        try:
            self.hdyn_h_h = self.hdyn_h_h.get()
            self.hdyn_d_h = self.hdyn_d_h.get()
            d_h_temp = d_h_temp.get()
            out = out.get()

        except AttributeError:
            pass

        if self.return_extracted_snr:
            return np.array([out, d_h_temp.real / np.sqrt(self.hdyn_h_h.real)]).T
        else:
            return out



# Useful MBHB-related Transformations

# Copyright (C) 2021 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from scipy import constants as ct

from .constants import *
from .citations import *

OrbitR = AU_SI


def mT_q(mT, q):
    """Convert total mass and mass ratio to m1,m2

    Args:
        mT (scalar or np.ndarray): Total mass of system.
        q (scalar or np.ndarray): Mass ratio of system with :math:`q<1`.

    Returns:
        tuple: First entry is ``m1``. Second entry is ``m2``.


    """
    return (mT / (1 + q), mT * q / (1 + q))


def modpi(phase):
    """Modulus with pi as the period

    Originally from Sylvain Marsat.

    Args:
        phase (scalar or np.ndarray): Phase angle.

    Returns:
        scalar or np.ndarray: Phase angle modulus by pi.

    """
    # from sylvain
    return phase - np.floor(phase / np.pi) * np.pi


def mod2pi(phase):
    """Modulus with 2pi as the period

    Originally from Sylvain Marsat.

    Args:
        phase (scalar or np.ndarray): Phase angle.

    Returns:
        scalar or np.ndarray: Phase angle modulus by 2pi.

    """
    # from sylvain
    return phase - np.floor(phase / (2 * np.pi)) * 2 * np.pi


def tSSBfromLframe(tL, lambdaSSB, betaSSB, t0=0.0):
    """Get time in SSB frame from time in LISA-frame.

    Compute Solar System Barycenter time ``tSSB`` from retarded time at the center
    of the LISA constellation ``tL``. **NOTE**: depends on the sky position
    given in solar system barycenter (SSB) frame.

    Originally from Sylvain Marsat. For more information and citation, see
    `arXiv:2003.00357 <https://arxiv.org/abs/2003.00357>`_.

    Args:
        tL (scalar or np.ndarray): Time in LISA constellation reference frame.
        lambdaSSB (scalar or np.ndarray): Ecliptic longitude in
            SSB reference frame.
        betaSSB (scalar or np.ndarray): Ecliptic latitude in SSB reference frame.
        t0 (double, optional): Initial start time point away from zero.
            (Default: ``0.0``)

    Returns:
        scalar or np.ndarray: Time in the SSB frame.

    """
    ConstPhi0 = ConstOmega * t0
    phase = ConstOmega * tL + ConstPhi0 - lambdaSSB
    RoC = OrbitR / C_SI
    return (
        tL
        + RoC * np.cos(betaSSB) * np.cos(phase)
        - 1.0 / 2 * ConstOmega * pow(RoC * np.cos(betaSSB), 2) * np.sin(2.0 * phase)
    )


# Compute retarded time at the center of the LISA constellation tL from Solar System Barycenter time tSSB */
def tLfromSSBframe(tSSB, lambdaSSB, betaSSB, t0=0.0):
    """Get time in LISA frame from time in SSB-frame.

    Compute retarded time at the center of the LISA constellation frame ``tL`` from
    the time in the SSB frame ``tSSB``. **NOTE**: depends on the sky position
    given in solar system barycenter (SSB) frame.

    Originally from Sylvain Marsat. For more information and citation, see
    `arXiv:2003.00357 <https://arxiv.org/abs/2003.00357>`_.

    Args:
        tSSB (scalar or np.ndarray): Time in LISA constellation reference frame.
        lambdaSSB (scalar or np.ndarray): Ecliptic longitude in
            SSB reference frame.
        betaSSB (scalar or np.ndarray): Time in LISA constellation reference frame.
        t0 (double, optional): Initial start time point away from zero.
            (Default: ``0.0``)

    Returns:
        scalar or np.ndarray: Time in the LISA frame.

    """
    ConstPhi0 = ConstOmega * t0
    phase = ConstOmega * tSSB + ConstPhi0 - lambdaSSB
    RoC = OrbitR / C_SI
    return tSSB - RoC * np.cos(betaSSB) * np.cos(phase)


def LISA_to_SSB(tL, lambdaL, betaL, psiL, t0=0.0):
    """Convert sky/orientation from LISA frame to SSB frame.

    Convert the sky and orientation parameters from the center of the LISA
    constellation reference to the SSB reference frame.

    The parameters that are converted are the reference time, ecliptic latitude,
    ecliptic longitude, and polarization angle.

    Originally from Sylvain Marsat. For more information and citation, see
    `arXiv:2003.00357 <https://arxiv.org/abs/2003.00357>`_.

    Args:
        tL (scalar or np.ndarray): Time in LISA constellation reference frame.
        lambdaL (scalar or np.ndarray): Ecliptic longitude in
            LISA reference frame.
        betaL (scalar or np.ndarray): Ecliptic latitude in LISA reference frame.
        psiL (scalar or np.ndarray): Polarization angle in LISA reference frame.
        t0 (double, optional): Initial start time point away from zero.
            (Default: ``0.0``)

    Returns:
        Tuple: (``tSSB``, ``lambdaSSB``, ``betaSSB``, ``psiSSB``)


    """

    t0 = t0 * YRSID_SI

    ConstPhi0 = ConstOmega * t0
    coszeta = np.cos(np.pi / 3.0)
    sinzeta = np.sin(np.pi / 3.0)
    coslambdaL = np.cos(lambdaL)
    sinlambdaL = np.sin(lambdaL)
    cosbetaL = np.cos(betaL)
    sinbetaL = np.sin(betaL)
    cospsiL = np.cos(psiL)
    sinpsiL = np.sin(psiL)
    lambdaSSB_approx = 0.0
    betaSSB_approx = 0.0
    # Initially, approximate alpha using tL instead of tSSB - then iterate */
    tSSB_approx = tL
    for k in range(3):
        alpha = ConstOmega * tSSB_approx + ConstPhi0
        cosalpha = np.cos(alpha)
        sinalpha = np.sin(alpha)
        lambdaSSB_approx = np.arctan2(
            cosalpha * cosalpha * cosbetaL * sinlambdaL
            - sinalpha * sinbetaL * sinzeta
            + cosbetaL * coszeta * sinalpha * sinalpha * sinlambdaL
            - cosalpha * cosbetaL * coslambdaL * sinalpha
            + cosalpha * cosbetaL * coszeta * coslambdaL * sinalpha,
            cosbetaL * coslambdaL * sinalpha * sinalpha
            - cosalpha * sinbetaL * sinzeta
            + cosalpha * cosalpha * cosbetaL * coszeta * coslambdaL
            - cosalpha * cosbetaL * sinalpha * sinlambdaL
            + cosalpha * cosbetaL * coszeta * sinalpha * sinlambdaL,
        )
        betaSSB_approx = np.arcsin(
            coszeta * sinbetaL
            + cosalpha * cosbetaL * coslambdaL * sinzeta
            + cosbetaL * sinalpha * sinzeta * sinlambdaL
        )
        tSSB_approx = tSSBfromLframe(tL, lambdaSSB_approx, betaSSB_approx, t0)

    lambdaSSB_approx = lambdaSSB_approx % (2 * np.pi)
    #  /* Polarization */
    psiSSB = modpi(
        psiL
        + np.arctan2(
            cosalpha * sinzeta * sinlambdaL - coslambdaL * sinalpha * sinzeta,
            cosbetaL * coszeta
            - cosalpha * coslambdaL * sinbetaL * sinzeta
            - sinalpha * sinbetaL * sinzeta * sinlambdaL,
        )
    )

    return (tSSB_approx, lambdaSSB_approx, betaSSB_approx, psiSSB)


def SSB_to_LISA(tSSB, lambdaSSB, betaSSB, psiSSB, t0=0.0):
    """Convert sky/orientation from SSB frame to LISA frame.

    Convert the sky and orientation parameters from the SSB reference frame to the center of the LISA
    constellation reference frame.

    The parameters that are converted are the reference time, ecliptic latitude,
    ecliptic longitude, and polarization angle.

    Originally from Sylvain Marsat. For more information and citation, see
    `arXiv:2003.00357 <https://arxiv.org/abs/2003.00357>`_.
    **Note**: no transformation of the phase -- approximant-dependence.

    Args:
        tSSB (scalar or np.ndarray): Time in SSB reference frame.
        lambdaSSB (scalar or np.ndarray): Ecliptic longitude in
            SSB reference frame.
        betaSSB (scalar or np.ndarray): Ecliptic latitude in SSB reference frame.
        psiSSB (scalar or np.ndarray): Polarization angle in SSB reference frame.
        t0 (double, optional): Initial start time point away from zero in years.
            (Default: ``0.0``)

    Returns:
        Tuple: (``tL``, ``lambdaL``, ``betaL``, ``psiL``)

    """
    t0 = t0 * YRSID_SI

    ConstPhi0 = ConstOmega * t0
    alpha = 0.0
    cosalpha = 0
    sinalpha = 0.0
    coslambda = 0
    sinlambda = 0.0
    cosbeta = 0.0
    sinbeta = 0.0
    cospsi = 0.0
    sinpsi = 0.0
    coszeta = np.cos(np.pi / 3.0)
    sinzeta = np.sin(np.pi / 3.0)
    coslambda = np.cos(lambdaSSB)
    sinlambda = np.sin(lambdaSSB)
    cosbeta = np.cos(betaSSB)
    sinbeta = np.sin(betaSSB)
    cospsi = np.cos(psiSSB)
    sinpsi = np.sin(psiSSB)
    alpha = ConstOmega * tSSB + ConstPhi0
    cosalpha = np.cos(alpha)
    sinalpha = np.sin(alpha)
    tL = tLfromSSBframe(tSSB, lambdaSSB, betaSSB, t0)
    lambdaL = np.arctan2(
        cosalpha * cosalpha * cosbeta * sinlambda
        + sinalpha * sinbeta * sinzeta
        + cosbeta * coszeta * sinalpha * sinalpha * sinlambda
        - cosalpha * cosbeta * coslambda * sinalpha
        + cosalpha * cosbeta * coszeta * coslambda * sinalpha,
        cosalpha * sinbeta * sinzeta
        + cosbeta * coslambda * sinalpha * sinalpha
        + cosalpha * cosalpha * cosbeta * coszeta * coslambda
        - cosalpha * cosbeta * sinalpha * sinlambda
        + cosalpha * cosbeta * coszeta * sinalpha * sinlambda,
    )
    betaL = np.arcsin(
        coszeta * sinbeta
        - cosalpha * cosbeta * coslambda * sinzeta
        - cosbeta * sinalpha * sinzeta * sinlambda
    )
    psiL = modpi(
        psiSSB
        + np.arctan2(
            coslambda * sinalpha * sinzeta - cosalpha * sinzeta * sinlambda,
            cosbeta * coszeta
            + cosalpha * coslambda * sinbeta * sinzeta
            + sinalpha * sinbeta * sinzeta * sinlambda,
        )
    )

    return (tL, lambdaL, betaL, psiL)


def mbh_sky_mode_transform(
    coords, ind_map=None, kind="both", inplace=False, cos_i=False
):
    """Sky-mode transformation in the LISA reference frame

    In the LISA constellation referenence frame, the sky localization for a
    source generally has 8 sky modes: 4 longitudinal and 2 latitudinal.
    Longitudinal modes involve :math:`\lambda + (0, 1, 2, 3)\\times\pi/2` and
    :math:`\psi + (0, 1, 2, 3)\\times\pi/2`, where :math:`\lambda` is the
    ecliptic longitude and :math:`\psi` is the polarization angle. The
    latitudinal modes are at :math:`(\pm\\beta, \pm\cos{\iota}, \pm\cos{\psi})`
    where :math:`\\beta` and :math:`\iota` are the ecliptic latitude and inclination,
    respectively.

    It is generally wise to start MBH PE runs in all eight sky modes. Using
    a mode-hopping proposal is also a good idea and/or parallel tempering.
    For MBHBs at higher frequency, the 8 modes reduce only to the latitudinal
    modes located at the true ecliptic longitude.

    Args:
        coords (np.ndarray): 2D array with shape: ``(num_bin_all, ndim)``.
            ``num_bin_all`` is the total number of binaries. ``ndim`` is the
            number of parameters given. If the inclination is given as :math:`\cos{\iota}`,
            that needs to be indicated with ``cos_i=True``. It does not matter if
            :math:`\\beta` or :math:`\sin{\\beta}` is given because the transforms
            involve multiplying the input by -1.
        ind_map (dict, optional): Keys are parameter names and values are there index
            into the coords array. Must include the keys ``inc, lam, beta, psi``.
            If ``None``, it will be ``dict(inc=7, lam=8, beta=9, psi=10)``.
            (Default: ``None``)
        kind (str, optional): String indicating which transform is needed.
            If ``kind=="both"`` (``kind=="long") [``kind=="lat"``],
            the transformation will be to all (longitudinal) [latitudinal] sky modes.
            If ``inplace==False``, the output array 1st dimension will be a factor
            of 8 (4) [2] longer. (Default: ``"both"``)
        inplace (bool, optional): If ``True``, adjust ``coords`` in place. In this
            case, the user needs to ensure the number of sets of binary Parameters
            is an integer multiple of the number of sky modes associated with the
            trasnformation. If ``kind=="both"`` (``kind=="long") [``kind=="lat"``],
            this must be a mutliple of 8 (4) [2].
            (Default: ``False``)
        cos_i (bool, optional): If ``True``, the inclination is input
            as the cosine of the inclination.
            (Default: ``False``)

    Returns:
        np.ndarray: 2D array with shape: ``(num_bin_all * factor, ndim)``.
            The factor is 1 if ``inplace==False``. The factor is 8 (4) [2]
            if ``kind=="both"`` (``kind=="long") [``kind=="lat"``].

    Raises:
        ValueError: Input arguments are not properly given.

    """

    # initialize ind_map
    if ind_map is None:
        ind_map = dict(inc=7, lam=8, beta=9, psi=10)

    elif isinstance(ind_map, dict) is False:
        raise ValueError("If providing the ind_map kwarg, it must be a dict.")

    if kind not in ["both", "lat", "long"]:
        raise ValueError(
            "The kwarg 'kind' must be lat for latitudinal transformation, long for longitudinal transformation, or both for both."
        )

    # factors for transformations
    elif kind == "both":
        factor = 8

    elif kind == "long":
        factor = 4

    elif kind == "lat":
        factor = 2

    if inplace:
        if (coords.shape[0] % factor) != 0:
            raise ValueError(
                "If performing an inplace transformation, the coords provided must have a first dimension size divisible by {} for a '{}' transformation.".format(
                    factor, kind
                )
            )

    else:
        coords = np.tile(coords, (factor, 1))

    if kind == "both" or kind == "lat":

        # inclination
        if cos_i:
            coords[1::2, ind_map["inc"]] *= -1

        else:
            coords[1::2, ind_map["inc"]] = np.pi - coords[1::2, ind_map["inc"]]

        # beta
        coords[1::2, ind_map["beta"]] *= -1

        # psi
        coords[1::2, ind_map["psi"]] = np.pi - coords[1::2, ind_map["psi"]]

    if kind == "long":
        for i in range(1, 4):
            # lambda
            coords[i::4, ind_map["lam"]] = (
                coords[i::4, ind_map["lam"]] + np.pi / 2 * i
            ) % (2 * np.pi)

            # psi
            coords[i::4, ind_map["psi"]] = (
                coords[i::4, ind_map["psi"]] + np.pi / 2 * i
            ) % (np.pi)

    if kind == "both":
        # indexing needs to be different in this case for longitudinal transforms
        num = coords.shape[0]
        for i in range(1, 4):
            for j in range(2):
                # lambda
                coords[2 * i + j :: 8, ind_map["lam"]] = (
                    coords[2 * i + j :: 8, ind_map["lam"]] + np.pi / 2 * i
                ) % (2 * np.pi)

                # psi
                coords[2 * i + j :: 8, ind_map["psi"]] = (
                    coords[2 * i + j :: 8, ind_map["psi"]] + np.pi / 2 * i
                ) % (np.pi)

    return coords



import numpy as np
cimport numpy as np
from libcpp cimport bool

from bbhx.utils.utility import pointer_adjust

assert sizeof(int) == sizeof(np.int32_t)

cdef extern from "Detector.hpp":
    ctypedef void* Orbits 'Orbits'

cdef extern from "Response.hh":
    ctypedef void* cmplx 'cmplx'

    void LISA_response(
        double* response_out,
        int* ells_in,
        int* mms_in,
        double* freqs,
        double* phi_ref,
        double* inc,
        double* lam,
        double* beta,
        double* psi,
        int TDItag, bool rescaled, bool tdi2, int order_fresnel_stencil,
        int numModes,
        int length,
        int numBinAll,
        int includesAmps,
        Orbits *orbits
    );

@pointer_adjust
def LISA_response_wrap(
     response_out,
     ells,
     mms,
     freqs,
     phi_ref,
     inc,
     lam,
     beta,
     psi,
    TDItag, rescaled, tdi2, order_fresnel_stencil,
    numModes,
    length,
    numBinAll,
    includesAmps,
    orbits
):

    cdef size_t response_out_in = response_out
    cdef size_t ells_in = ells
    cdef size_t mms_in = mms
    cdef size_t freqs_in = freqs
    cdef size_t inc_in = inc
    cdef size_t lam_in = lam
    cdef size_t beta_in = beta
    cdef size_t psi_in = psi
    cdef size_t phi_ref_in = phi_ref
    cdef size_t orbits_in = orbits

    LISA_response(
        <double*> response_out_in,
        <int*> ells_in,
        <int*> mms_in,
        <double*> freqs_in,
        <double*> phi_ref_in,
        <double*> inc_in,
        <double*> lam_in,
        <double*> beta_in,
        <double*> psi_in,
        TDItag, rescaled, tdi2, order_fresnel_stencil,
        numModes,
        length,
        numBinAll,
        includesAmps,
        <Orbits*> orbits_in
    )



import numpy as np
cimport numpy as np
from libcpp cimport bool

from bbhx.utils.utility import pointer_adjust

assert sizeof(int) == sizeof(np.int32_t)

cdef extern from "Likelihood.hh":
    ctypedef void* cmplx 'cmplx'

    void hdyn(cmplx* likeOut1, cmplx* likeOut2,
                        cmplx* templateChannels, cmplx* dataConstants,
                        double* dataFreqs,
                        int numBinAll, int data_length, int nChannels);

    void direct_like(cmplx* d_h, cmplx* h_h, cmplx* dataChannels, double* noise_weight_times_df, long* templateChannels_ptrs, int* inds_start, int* ind_lengths, int data_stream_length, int numBinAll);

    void prep_hdyn_wrap(cmplx* A0_in, cmplx* A1_in, cmplx* B0_in, cmplx* B1_in, cmplx* d_arr, cmplx* h0_arr, double* S_n_arr, double df, int* bins, double* f_dense, double* f_m_arr, int data_length, int nchannels, int length_f_rel);

@pointer_adjust
def hdyn_wrap(likeOut1, likeOut2,
                    templateChannels, dataConstants,
                    dataFreqs,
                    numBinAll, data_length, nChannels):

    cdef size_t likeOut1_in = likeOut1
    cdef size_t likeOut2_in = likeOut2
    cdef size_t templateChannels_in = templateChannels
    cdef size_t dataConstants_in = dataConstants
    cdef size_t dataFreqs_in = dataFreqs

    hdyn(<cmplx*> likeOut1_in, <cmplx*> likeOut2_in,
            <cmplx*> templateChannels_in, <cmplx*> dataConstants_in,
            <double*> dataFreqs_in,
            numBinAll, data_length, nChannels);

@pointer_adjust
def direct_like_wrap(d_h, h_h, dataChannels, noise_weight_times_df, templateChannels_ptrs, inds_start, ind_lengths, data_stream_length, numBinAll):

    cdef size_t d_h_in = d_h
    cdef size_t h_h_in = h_h
    cdef size_t dataChannels_in = dataChannels
    cdef size_t noise_weight_times_df_in = noise_weight_times_df
    cdef size_t templateChannels_ptrs_in = templateChannels_ptrs
    cdef size_t inds_start_in = inds_start
    cdef size_t ind_lengths_in = ind_lengths

    direct_like(<cmplx*> d_h_in, <cmplx*> h_h_in, <cmplx*> dataChannels_in, <double*> noise_weight_times_df_in, <long*> templateChannels_ptrs_in, <int*> inds_start_in, <int*> ind_lengths_in, data_stream_length, numBinAll)


@pointer_adjust
def prep_hdyn(A0_in, A1_in, B0_in, B1_in, d_arr, h0_arr, S_n_arr, df, bins, f_dense, f_m_arr, data_length, nchannels, length_f_rel):

    cdef size_t A0_in_in = A0_in
    cdef size_t A1_in_in = A1_in
    cdef size_t B0_in_in = B0_in
    cdef size_t B1_in_in = B1_in
    cdef size_t d_arr_in = d_arr
    cdef size_t h0_arr_in = h0_arr
    cdef size_t S_n_arr_in = S_n_arr
    cdef size_t f_dense_in = f_dense
    cdef size_t f_m_arr_in = f_m_arr
    cdef size_t bins_in = bins

    prep_hdyn_wrap(<cmplx*> A0_in_in, <cmplx*> A1_in_in, <cmplx*> B0_in_in, <cmplx*> B1_in_in, <cmplx*> d_arr_in, <cmplx*> h0_arr_in, <double*> S_n_arr_in, df, <int*> bins_in, <double*> f_dense_in, <double*> f_m_arr_in, data_length, nchannels, length_f_rel)



import numpy as np
cimport numpy as np

from bbhx.utils.utility import pointer_adjust

assert sizeof(int) == sizeof(np.int32_t)

cdef extern from "PhenomHM.hh":
    void waveform_amp_phase(
        double* waveformOut,
        int* ells_in,
        int* mms_in,
        double* freqs,
        double* m1_SI,
        double* m2_SI,
        double* chi1z,
        double* chi2z,
        double* distance,
        double* f_ref,
        int numModes,
        int length,
        int numBinAll,
        double* Mf_RD_lm_all,
        double* Mf_DM_lm_all,
        int run_phenomd
    )

    void get_phenomhm_ringdown_frequencies_wrap(
        double *fringdown,
        double *fdamp,
        double *m1,
        double *m2,
        double *chi1z,
        double *chi2z,
        int *ells_in,
        int *mm_in,
        int numModes,
        int numBinAll
    )

    void get_phenomd_ringdown_frequencies_wrap(
        double *fringdown,
        double *fdamp,
        double *m1,
        double *m2,
        double *chi1z,
        double *chi2z,
        int numBinAll,
        double *y_rd_all,
        double *c1_rd_all,
        double *c2_rd_all,
        double *c3_rd_all,
        double *y_dm_all,
        double *c1_dm_all,
        double *c2_dm_all,
        double *c3_dm_all,
        double dspin
    );

@pointer_adjust
def waveform_amp_phase_wrap(
    waveformOut,
    ells,
    mms,
    freqs,
    m1_SI,
    m2_SI,
    chi1z,
    chi2z,
    distance,
    f_ref,
    numModes,
    length,
    numBinAll,
    Mf_RD_lm_all,
    Mf_DM_lm_all,
    run_phenomd
):

    cdef size_t waveformOut_in = waveformOut
    cdef size_t ells_in = ells
    cdef size_t mms_in = mms
    cdef size_t freqs_in = freqs
    cdef size_t m1_SI_in = m1_SI
    cdef size_t m2_SI_in = m2_SI
    cdef size_t chi1z_in = chi1z
    cdef size_t chi2z_in = chi2z
    cdef size_t distance_in = distance
    cdef size_t f_ref_in = f_ref
    cdef size_t Mf_RD_lm_all_in = Mf_RD_lm_all
    cdef size_t Mf_DM_lm_all_in = Mf_DM_lm_all


    waveform_amp_phase(
        <double*> waveformOut_in,
        <int*> ells_in,
        <int*> mms_in,
        <double*> freqs_in,
        <double*> m1_SI_in,
        <double*> m2_SI_in,
        <double*> chi1z_in,
        <double*> chi2z_in,
        <double*> distance_in,
        <double*> f_ref_in,
        numModes,
        length,
        numBinAll,
        <double*> Mf_RD_lm_all_in,
        <double*> Mf_DM_lm_all_in,
        run_phenomd
    )

    return

@pointer_adjust
def get_phenomhm_ringdown_frequencies(
    fringdown,
    fdamp,
    m1,
    m2,
    chi1z,
    chi2z,
    ells,
    mms,
    numModes,
    numBinAll,
):

    cdef size_t fringdown_in = fringdown
    cdef size_t fdamp_in = fdamp
    cdef size_t m1_in = m1
    cdef size_t m2_in = m2
    cdef size_t chi1z_in = chi1z
    cdef size_t chi2z_in = chi2z
    cdef size_t ells_in = ells
    cdef size_t mms_in = mms

    get_phenomhm_ringdown_frequencies_wrap(
        <double *>fringdown_in,
        <double *>fdamp_in,
        <double *>m1_in,
        <double *>m2_in,
        <double *>chi1z_in,
        <double *> chi2z_in,
        <int *> ells_in,
        <int *> mms_in,
        numModes,
        numBinAll
    )
    return

@pointer_adjust
def get_phenomd_ringdown_frequencies(
    fringdown,
    fdamp,
    m1,
    m2,
    chi1z,
    chi2z,
    numBinAll,
    y_rd_all,
    c1_rd_all,
    c2_rd_all,
    c3_rd_all,
    y_dm_all,
    c1_dm_all,
    c2_dm_all,
    c3_dm_all,
    dspin
):

    cdef size_t fringdown_in = fringdown
    cdef size_t fdamp_in = fdamp
    cdef size_t m1_in = m1
    cdef size_t m2_in = m2
    cdef size_t chi1z_in = chi1z
    cdef size_t chi2z_in = chi2z
    cdef size_t y_rd_all_in = y_rd_all
    cdef size_t c1_rd_all_in = c1_rd_all
    cdef size_t c2_rd_all_in = c2_rd_all
    cdef size_t c3_rd_all_in = c3_rd_all
    cdef size_t y_dm_all_in = y_dm_all
    cdef size_t c1_dm_all_in = c1_dm_all
    cdef size_t c2_dm_all_in = c2_dm_all
    cdef size_t c3_dm_all_in = c3_dm_all



    get_phenomd_ringdown_frequencies_wrap(
        <double*> fringdown_in,
        <double*> fdamp_in,
        <double*> m1_in,
        <double*> m2_in,
        <double*> chi1z_in,
        <double*> chi2z_in,
        numBinAll,
        <double*> y_rd_all_in,
        <double*> c1_rd_all_in,
        <double*> c2_rd_all_in,
        <double*> c3_rd_all_in,
        <double*> y_dm_all_in,
        <double*> c1_dm_all_in,
        <double*> c2_dm_all_in,
        <double*> c3_dm_all_in,
        dspin
    )
    return
